// 8to16.cpp : Defines the entry point for the console application.
//
//	This program reads an 8-bit color Windows Bitmap (argv[1]) and
//	writes a 16-bit pseudo-Windows Bitmap (argv[2])
//
//	The output file is suitable for start-up screens on some 
//  Honda/Acura navigation systems

#include "stdafx.h"
#include <windows.h>

typedef unsigned char	U8,  *PU8;
typedef unsigned short	U16, *PU16;
typedef unsigned long	U32, *PU32;


typedef struct
	{
	char	identifier[2];		// 'BM' usually
	U32		filesize;			// 113398
	U32		reserved;
	U32		bitmapDataOffset;	// 1078
	U32		bitmapHeaderSize;	// 40
	U32		width;				// 480
	U32		height;				// 234
	U16		planes;				// 1
	U16		bitsPerPixel;		// 8
	U32		compression;		// 0
	U32		bitmapDataSize;		// 112320
	U32		hres;				// d/c
	U32		vres;				// d/c
	U32		colors;				// 256
	U32		importantColors;	// 256
	} BMP40, *PBMP40;

BMP40	bmp40;
U32		palette[256];
U8		data[234][480];
U32		filler[7];

int main(int argc, char* argv[])
{
	HANDLE	h_in, h_out;
	U32		dwBytes, x, y;
	U32		input;
	U16		output;
	U8		r, g, b;

	// there must be two arguments
	if (argc != 3)
		{
		fprintf(stderr, "?Usage: %s filename_in.bmp filename_out.bmp\n", argv[0]);
		return 1;
		}

	// open the input file
	h_in = CreateFile(argv[1], GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (h_in == INVALID_HANDLE_VALUE)
		{
		fprintf(stderr, "?Error %u Opening '%s'.\n", GetLastError(), argv[1]);
		return 2;
		}

	// warn if file is not the expected size.  If not, the rest of the program is not likely to work.
	if (GetFileSize(h_in, NULL) != 113398L)
		{
		fprintf(stderr, "?Surprise: Input file '%s' is not exactly 113,398 bytes.\n", argv[2]);
//		return 3;
		}

	// open the output file
	h_out = CreateFile(argv[2], GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (h_out == INVALID_HANDLE_VALUE)
		{
		fprintf(stderr, "?Error %u Creating '%s'.\n", GetLastError(), argv[2]);
		return 4;
		}

	// read the standard Windows Bitmap file header
	if (!ReadFile(h_in, (void *)&bmp40, sizeof(bmp40), &dwBytes, NULL) || (dwBytes != sizeof(bmp40)))
		{
		fprintf(stderr, "?Error %u Reading bitmap header from '%s'.\n", GetLastError(), argv[1]);
		return 5;
		}

	// get offset to palette and show it
	dwBytes = SetFilePointer(h_in, 0, NULL, FILE_CURRENT);
	printf("palette starts at %08x\n", dwBytes);

	// read palette, should be 256 32-bit entries
	if (!ReadFile(h_in, (void *)palette, sizeof(palette), &dwBytes, NULL) || (dwBytes != sizeof(palette)))
		{
		fprintf(stderr, "?Error %u Reading palette from '%s'.\n", GetLastError(), argv[1]);
		return 6;
		}

	// get offset to data and show it
	dwBytes = SetFilePointer(h_in, 0, NULL, FILE_CURRENT);
	printf("data starts at %08x\n", dwBytes);

	// read data, should be 480 * 234 * 8-bits.
	if (!ReadFile(h_in, (void *)data, sizeof(data), &dwBytes, NULL) || (dwBytes != sizeof(data)))
		{
		fprintf(stderr, "?Error %u Reading data from '%s'. Read %i bytes.\n", GetLastError(), argv[1], dwBytes);
		return 7;
		}

	// we're one with the input file
	CloseHandle(h_in);

	// make header similar to what Honda's Opening1.bmp has.
	bmp40.filesize = 224712;
	bmp40.bitmapDataOffset = 82;
	bmp40.bitmapHeaderSize = 56;
	bmp40.bitsPerPixel = 16;
	bmp40.compression = 3;
	bmp40.bitmapDataSize = 0;
	bmp40.hres = 2834;
	bmp40.vres = 2834;
	bmp40.colors = 0;
	bmp40.importantColors = 0;

	// write header to output file
	if (!WriteFile(h_out, (void *)&bmp40, sizeof(bmp40), &dwBytes, NULL) || (dwBytes != sizeof(bmp40)))
		{
		fprintf(stderr, "?Error %u Writing header to '%s'.\n", GetLastError(), argv[2]);
		return 10;
		}

	// initialize some of the 28 extra bytes the new header has.
	filler[0] = 0x00007c00;
	filler[1] = 0x000003e0;
	filler[2] = 0x0000001f;
	filler[3] = 0x00008000;

	// write the 28 extra bytes to the output file
	if (!WriteFile(h_out, (void *)filler, sizeof(filler), &dwBytes, NULL) || (dwBytes != sizeof(filler)))
		{
		fprintf(stderr, "?Error %u Writing filler to '%s'.\n", GetLastError(), argv[2]);
		return 11;
		}

	// set through each row and column, converting the 32-bit color in the palette to
	// a 16-bit color.
	// input color looks like: 00000000RRRRRrrrGGGGGgggBBBBBbbb
	// output color looks like:0RRRRRGGGGGBBBBB
	for (y=0; y<234; y++)
		for (x=0; x<480; x++)
			{
			// find the color in the palette
			input = palette[data[y][x]];

			// break color up into bytes
			r = (U8)(input >> 16);	// I might have the order reversed, but it doesn't matter
			g = (U8)(input >> 8);
			b = (U8)(input >> 0);

			// string together highest 5 biyts in each byte
			output = ((r >> 3) << 10) | ((g >> 3) << 5) | ((b >> 3) << 0);

			// write 16-bit result to output file
			if (!WriteFile(h_out, (void *)&output, sizeof(output), &dwBytes, NULL) || (dwBytes != sizeof(output)))
				{
				fprintf(stderr, "?Error %u Writing output to '%s'.\n", GetLastError(), argv[2]);
				return 12;
				}
			}

	// We're done
	CloseHandle(h_out);

	return 0;
}

